'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol ? "symbol" : typeof obj; };

exports.mapType = mapType;
exports.toGraphQLDefinition = toGraphQLDefinition;
exports.attributeToGraphQLType = attributeToGraphQLType;

var _graphql = require('graphql');

var _graphqlRelay = require('graphql-relay');

var _thinky = require('thinky');

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var customTypeMapper = void 0;

/**
 * A function to set a custom mapping of types
 * @param {Function} mapFunc
 */
function mapType(mapFunc) {
  customTypeMapper = mapFunc;
}

/**
 * Create a graphql definition based to a
 * thinky model
 *
 * @param thinkyModel
 * @returns {*}
 */
function toGraphQLDefinition(thinkyModel) {
  var opts = arguments.length <= 1 || arguments[1] === undefined ? {} : arguments[1];

  opts.exclude = opts.exclude || [];
  opts.only = opts.only || null;
  opts.map = opts.map || {};
  opts.globalId = opts.globalId === undefined ? false : opts.globalId;
  opts.allowNull = opts.allowNull === undefined ? true : opts.allowNull;

  var modelSchema = thinkyModel._schema._schema,
      modelDef = thinkyModel._schema._model,
      modelName = modelDef._name;

  // Generate the GraphQL Definition Object
  var graphQLDefinition = Object.keys(modelSchema).reduce(function (gqlDef, key) {
    // Skip Excluded
    if (opts.exclude.indexOf(key) !== -1) {
      return gqlDef;
    }

    // Only those fields
    if (opts.only && opts.only.indexOf(key) === -1) {
      return gqlDef;
    }

    var modelType = modelSchema[key];
    var name = _thinky.type.isObject(modelType) ? _lodash2.default.upperFirst(modelDef._name) + _lodash2.default.upperFirst(key) : key;
    var graphQLType = attributeToGraphQLType(modelType, name);

    if (graphQLType) {
      // Map the field if specified
      if (opts.map.hasOwnProperty(key)) {
        if (typeof opts.map === 'function') {
          key = opts.map(key);
        } else {
          key = opts.map[key];
        }
      }

      // If the attribute is a "not nullable" field, then i'll wrap it
      // with a GraphQLNotNull instance
      if (opts.allowNull && modelType._options.enforce_type === 'strict') {
        gqlDef[key] = {
          type: new _graphql.GraphQLNonNull(graphQLType)
        };
      } else {
        gqlDef[key] = {
          type: graphQLType
        };
      }
    }

    return gqlDef;
  }, {});

  // Add relay global id if requested
  if (opts.globalId) {
    var field = modelName.toLowerCase() + 'ID';
    graphQLDefinition.id = (0, _graphqlRelay.globalIdField)(modelName, function (instance) {
      return instance.id;
    });
    graphQLDefinition[field] = {
      type: _graphql.GraphQLString,
      resolve: function resolve(source) {
        return source.id;
      }
    };
  } else {
    graphQLDefinition.id = {
      type: _graphql.GraphQLString
    };
  }

  return graphQLDefinition;
}

/**
 * Accept a thinky type and return
 * a graphQL Type
 *
 * @param attributeType
 * @param name
 */
function attributeToGraphQLType(attributeType, name) {
  var schema = attributeType._schema,
      schemaKeys = schema ? Object.keys(schema) : [],
      enumKeys = attributeType._enum ? Object.keys(attributeType._enum) : [];

  // User did create a custom type mapper
  // return it if it is truthy
  if (customTypeMapper) {
    var result = customTypeMapper(attributeType);
    if (result) {
      return result;
    }
  }

  // ENUM type
  if (enumKeys.length > 0) {
    var _ret = function () {
      var specialChars = /[^a-z\d_]/i;

      return {
        v: new _graphql.GraphQLEnumType({
          name: name + 'EnumType',
          values: enumKeys.reduce(function (obj, enumKey) {
            var value = enumKey;
            var sanitizedValue = value;
            if (specialChars.test(value)) {
              sanitizedValue = value.split(specialChars).reduce(function (reduced, val, idx) {
                var newVal = val;
                if (idx > 0) {
                  newVal = '' + val[0].toUpperCase() + val.slice(1);
                }
                return '' + reduced + newVal;
              });
            }
            obj[sanitizedValue] = { value: value };
            return obj;
          }, {})
        })
      };
    }();

    if ((typeof _ret === 'undefined' ? 'undefined' : _typeof(_ret)) === "object") return _ret.v;
  }

  // Bolean Type
  if (_thinky.type.isBoolean(attributeType)) {
    return _graphql.GraphQLBoolean;
  }

  // String Type
  if (_thinky.type.isString(attributeType) || _thinky.type.isDate(attributeType) || _thinky.type.isPoint(attributeType)) {
    return _graphql.GraphQLString;
  }

  // Number Type TODO: what about if it is float?
  if (_thinky.type.isNumber(attributeType)) {
    return _graphql.GraphQLInt;
  }

  // Object type
  if (_thinky.type.isObject(attributeType)) {
    var _ret2 = function () {
      if (schemaKeys.length === 0) {
        console.info('Attrbute type object doesnt have a schema, skip...');
        return {
          v: void 0
        };
      }

      var graphQLTypeDef = {};
      schemaKeys.forEach(function (attribute) {
        graphQLTypeDef[attribute] = {};

        var graphQLType = attributeToGraphQLType(schema[attribute], name);

        if (graphQLType) {
          graphQLTypeDef[attribute]['type'] = graphQLType;
        }
      });

      if (!name) {
        throw new Error('Specify a name for the Object Attribute type');
      }

      return {
        v: new _graphql.GraphQLObjectType({
          name: _lodash2.default.upperFirst(name),
          fields: _extends({}, graphQLTypeDef)
        })
      };
    }();

    if ((typeof _ret2 === 'undefined' ? 'undefined' : _typeof(_ret2)) === "object") return _ret2.v;
  }

  // Array type
  if (_thinky.type.isArray(attributeType)) {
    var schemaArray = attributeType._schema;
    var arrayType = attributeToGraphQLType(schemaArray, name);
    return new _graphql.GraphQLList(arrayType);
  }
}